﻿//////////////////////////////////////////////
// BufferCast.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "Buffer.h"

/// Class ------------------------------------

namespace nkMemory
{
	template <typename T = unsigned char>
	class BufferCast
	{
		private :

			// Attributes
			// Base buffer
			union
			{
				Buffer _buffer = Buffer() ;

				struct
				{
					T* _ptr ;
					unsigned long long _size ;
				} _debugLayout ;
			} ;

		public :

			// Constructors
			BufferCast () = default ;
			
			BufferCast (unsigned long long size)
			:	_buffer (size * sizeof(T))
			{
				// Nothing to do
			}

			BufferCast (T* data, unsigned long long size)
			:	_buffer ((unsigned char*)data, size * sizeof(T))
			{
				// Nothing to do
			}

			BufferCast (const Buffer& buffer)
			:	_buffer (buffer)
			{
				// Nothing to do
			}

			BufferCast (Buffer&& buffer)
			:	_buffer (std::move(buffer))
			{
				// Nothing to do
			}

			BufferCast (const BufferCast& cast)
			:	_buffer (cast._buffer)
			{
				// Nothing to do
			}

			BufferCast (BufferCast&& cast)
			:	_buffer (std::move(cast._buffer))
			{
				// Nothing to do
			}

			// Destructor
			~BufferCast ()
			{
				_buffer.~Buffer() ;
			}

		public :

			// Getters
			T* getData () const
			{
				return (T*)_buffer.getData() ;
			}

			unsigned long long getSize () const
			{
				return _buffer.getSize() / sizeof(T) ;
			}

			bool empty () const
			{
				return !getSize() ;
			}

			T& front ()
			{
				return getData()[0] ;
			}

			const T& front () const
			{
				return getData()[0] ;
			}

			T& back ()
			{
				return getData()[getSize() - 1] ;
			}

			const T& back () const
			{
				return getData()[getSize() - 1] ;
			}

		public :

			// Management
			void clear ()
			{
				_buffer.clear() ;
			}

			template <typename U = T>
			typename std::enable_if_t<!std::is_class_v<U>, void>
			resize (unsigned long long size)
			{
				_buffer.resize(size * sizeof(T)) ;
			}

			template <typename U = T>
			typename std::enable_if_t<std::is_class_v<U>, void>
			resize (unsigned long long size)
			{
				// Keep size diff
				long long oldSize = (long long)getSize() ;
				long long sizeDiff = size - oldSize ;

				// Resize, copying data
				_buffer.resize(size * sizeof(T)) ;

				// If we need to fill in new instances
				T* data = getData() ;

				for (long long i = 0 ; i < sizeDiff ; ++i)
					new (&data[oldSize + i]) T () ;
			}

			template <typename U = T>
			typename std::enable_if_t<!std::is_class_v<U>, T&>
			append (const T& value)
			{
				unsigned long long size = getSize() ;
				resize(size + 1) ;

				T* data = getData() ;
				data[size] = value ;

				return data[size] ;
			}

			template <typename U = T>
			typename std::enable_if_t<std::is_class_v<U>, T&>
			append (const T& value)
			{
				unsigned long long size = getSize() ;
				resize(size + 1) ;

				T* data = getData() ;
				new (&data[size]) T (value) ;

				return data[size] ;
			}

			template <typename U = T>
			typename std::enable_if_t<std::is_class_v<U>, T&>
			append (T&& value)
			{
				unsigned long long size = getSize() ;
				resize(size + 1) ;

				T* data = getData() ;
				new (&data[size]) T (std::forward<T>(value)) ;

				return data[size] ;
			}

			BufferCast<T>& erase (unsigned long long index, unsigned long long count = 1ull)
			{
				// Transfer to buffer
				_buffer.erase(index * sizeof(T), count * sizeof(T)) ;

				return *this ;
			}

			Buffer relinquishBufferOwnership ()
			{
				return std::move(_buffer) ;
			}

			BufferDataDescriptor relinquishDataOwnership ()
			{
				return _buffer.relinquishDataOwnership() ;
			}

		public :

			// Operators
			T& operator[] (unsigned long long index)
			{
				return getData()[index] ;
			}

			const T& operator[] (unsigned long long index) const
			{
				return getData()[index] ;
			}
			
			BufferCast<T>& operator= (const BufferCast<T>& other)
			{
				_buffer = other._buffer ;

				return *this ;
			}

			BufferCast<T>& operator= (BufferCast<T>&& other)
			{
				_buffer = std::move(other._buffer) ;

				return *this ;
			}

		public :

			// Templated constructors
			BufferCast (const std::vector<T>& vec)
			:	_buffer (vec)
			{
				// Nothing to do
			}
	} ;
}

/// Exports ----------------------------------

namespace nkMemory
{
	// Unsigned
	template class DLL_MEMORY_EXPORT BufferCast<unsigned char> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned short> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned int> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned long long> ;

	// Signed
	template class DLL_MEMORY_EXPORT BufferCast<char> ;
	template class DLL_MEMORY_EXPORT BufferCast<short> ;
	template class DLL_MEMORY_EXPORT BufferCast<int> ;
	template class DLL_MEMORY_EXPORT BufferCast<long long> ;

	// Floating
	template class DLL_MEMORY_EXPORT BufferCast<float> ;
	template class DLL_MEMORY_EXPORT BufferCast<double> ;
}